from monopoly.banks.base import BankBase
from monopoly.config import CreditStatementConfig
from monopoly.constants import InternalBankNames, MetadataIdentifier, SharedPatterns


class ExampleBank(BankBase):
    """Dummy class to help with reading the example PDF statement"""

    credit_config = CreditStatementConfig(
        bank_name=InternalBankNames.EXAMPLE,
        statement_date_pattern=r"(\d{2}\-\d{2}\-\d{4})",
        prev_balance_pattern=(
            r"(?P<description>LAST MONTH'S BALANCE?)\s+"
            + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
        ),
        transaction_pattern=(
            r"(?P<transaction_date>\d+/\d+)\s*"
            + SharedPatterns.DESCRIPTION
            + SharedPatterns.AMOUNT_EXTENDED
        ),
    )

    identifiers = [
        MetadataIdentifier(
            creator="Adobe Acrobat 23.3", producer="Adobe Acrobat Pro (64-bit)"
        )
    ]



import logging
from dataclasses import Field, fields
from itertools import product
from typing import Type

from monopoly.constants import EncryptionIdentifier, MetadataIdentifier

from ..examples.example_bank import ExampleBank
from .base import BankBase
from .citibank import Citibank
from .dbs import Dbs
from .hsbc import Hsbc
from .maybank import Maybank
from .ocbc import Ocbc
from .standard_chartered import StandardChartered

banks: list[Type[BankBase]] = [
    Citibank,
    Dbs,
    ExampleBank,
    Hsbc,
    Maybank,
    Ocbc,
    StandardChartered,
]

logger = logging.getLogger(__name__)


class UnsupportedBankError(Exception):
    """Raised when a processor cannot be found for a specific bank"""


def detect_bank(
    metadata_items: list[EncryptionIdentifier | MetadataIdentifier],
) -> Type[BankBase] | None:
    """
    Reads the encryption metadata or actual metadata (if the PDF is not encrypted),
    and checks for a bank based on unique identifiers.
    """
    for bank in banks:
        if is_bank_identified(metadata_items, bank):
            return bank
    return None


def is_bank_identified(
    metadata_items: list[EncryptionIdentifier | MetadataIdentifier],
    bank: Type[BankBase],
) -> bool:
    """
    Checks if a bank is identified based on a list of metadata items.
    """
    for identifier, metadata in product(
        bank.identifiers, metadata_items
    ):  # type: ignore
        # Only compare matching identifier types
        if type(metadata) is not type(identifier):
            continue

        logger.debug(
            "Checking if PDF %s matches %s %s",
            metadata.__class__.__name__,
            bank.__name__,
            identifier,
        )

        if all(
            check_matching_field(field, metadata, identifier)
            for field in fields(metadata)
        ):
            logger.debug("Identified statement bank: %s", bank.__name__)
            return True

    return False


def check_matching_field(
    field: Field,
    metadata: EncryptionIdentifier | MetadataIdentifier,
    identifier: EncryptionIdentifier | MetadataIdentifier,
) -> bool:
    """
    Checks if a field in the metadata matches the corresponding identifier field.
    """
    field_value = getattr(metadata, field.name)
    identifier_value = getattr(identifier, field.name)

    # if identifier is not set, we assume a match
    # this means only the identifiers in the bank
    # class need to be matched.
    if not identifier_value:
        return True

    # allow for partial string matching
    # since '' in a string always is true
    # add another condition to filter out blank strings
    partial_string_match = (
        isinstance(field.type(), str) and identifier_value in field_value
    )

    # other types should match exactly
    full_match = identifier_value == field_value

    if any([partial_string_match, full_match]):
        logger.debug(
            "Match found for field `%s` : '%s' - '%s'",
            field.name,
            identifier_value,
            field_value,
        )
        return True

    return False



import re
from typing import Any, Optional

from pydantic import ConfigDict, SecretStr
from pydantic.dataclasses import dataclass
from pydantic_settings import BaseSettings, SettingsConfigDict

from monopoly.constants import BankNames, InternalBankNames


class PdfPasswords(BaseSettings):
    ocbc_pdf_passwords: list[SecretStr] = [SecretStr("")]
    citibank_pdf_passwords: list[SecretStr] = [SecretStr("")]
    standard_chartered_pdf_passwords: list[SecretStr] = [SecretStr("")]
    hsbc_pdf_passwords: list[SecretStr] = [SecretStr("")]

    model_config = SettingsConfigDict(env_file=".env", extra="allow")


class Passwords(BaseSettings):
    """
    Pydantic model that automatically populates variables from a .env file,
    or an environment variable called `passwords`.
    e.g. export PASSWORDS='{"OCBC_PDF_PASSWORDS": ...}'
    """

    passwords: PdfPasswords = PdfPasswords()


@dataclass(frozen=True)
class DateOrder:
    """
    Supported `dateparser` DATE_ORDER arguments can be found here:
    https://dateparser.readthedocs.io/en/latest/settings.html#date-order
    """

    date_order: str

    @property
    def settings(self):
        return {"DATE_ORDER": self.date_order}


# pylint: disable=too-many-instance-attributes
@dataclass(kw_only=True)
class StatementConfig:
    """
    Base configuration class storing configuration values for debit and
    credit card statements

    - `transaction_pattern` refers to the regex pattern used to capture transactions,
    where a pattern like:
        "(?P<transaction_date>\\d+/\\d+)\\s*"
        "(?P<description>.*?)\\s*"
        "(?P<amount>[\\d.,]+)$"
    is used to capture a transaction like:
        06/07 URBAN TRANSIT CO. SINGAPORE SG  1.38
    - `transaction_date_order` represents the datetime format that a specific bank uses
    for transactions. For example, "DMY" will parse 01/02/2024 as 1 Feb 2024.
    Defaults to DMY.
    - `multiline_transactions` controls whether Monopoly tries to concatenate
    transactions that are split across two lines
    - `debit_statement_identifier` is a regex pattern that is used to determine whether
    a statement from a bank is a debit or credit card statement.
    """

    bank_name: BankNames | InternalBankNames
    transaction_pattern: str
    statement_date_pattern: str
    transaction_date_order: DateOrder = DateOrder("DMY")
    statement_date_order: DateOrder = DateOrder("DMY")
    multiline_transactions: bool = False
    has_withdraw_deposit_column: bool = False


@dataclass(config=ConfigDict(extra="forbid"), kw_only=True)
class DebitStatementConfig(StatementConfig):
    """
    Dataclass storing configuration values unique to debit statements
    """

    has_withdraw_deposit_column: bool = True
    debit_statement_identifier: Optional[str] = None


@dataclass(config=ConfigDict(extra="forbid"))
class CreditStatementConfig(StatementConfig):
    """
    Dataclass storing configuration values unique to credit statements

    - `prev_balance_pattern` is a regex pattern used to match the previous balance
    line in a credit statements, which is then treated as a transaction.
    """

    prev_balance_pattern: Optional[Any | re.Pattern] = None

    def __post_init__(self):
        if self.prev_balance_pattern:
            self.prev_balance_pattern = re.compile(self.prev_balance_pattern)


@dataclass
class PdfConfig:
    """Stores PDF configuration values for the `PdfParser` class

    - `password`: The password used to unlock the PDF (if it is locked)
    - `page_range`: A slice representing which pages to process. For
    example, a range of (1, -1) will mean that the first and last pages
    are skipped.
    - `page_bbox`: A tuple representing the bounding box range for every
    page. This is used to avoid weirdness like vertical text, and other
    PDF artifacts that may affect parsing.
    """

    page_range: tuple[Optional[int], Optional[int]] = (None, None)
    page_bbox: Optional[tuple[float, float, float, float]] = None


passwords = Passwords().passwords



import logging
from dataclasses import dataclass
from functools import cached_property, lru_cache
from io import BytesIO
from pathlib import Path
from typing import Optional, Type

import fitz
import pdftotext
from pydantic import SecretStr

from monopoly.banks import BankBase

logger = logging.getLogger(__name__)


@dataclass
class PdfPage:
    """
    Dataclass representation of a bank statement PDF page.
    Contains the raw text of a PDF page, and allows access
    to the raw text as a list via the `lines` property
    """

    raw_text: str

    @cached_property
    def lines(self) -> list[str]:
        return self.raw_text.split("\n")


class WrongPasswordError(Exception):
    """Exception raised when an incorrect password is provided"""


class MissingPasswordError(Exception):
    """Exception raised when the document is encrypted, but no password is provided"""


class BadPasswordFormatError(Exception):
    """Exception raised passwords are not provided in a proper format"""


class PdfParser:
    def __init__(
        self,
        bank: Type[BankBase],
        passwords: Optional[list[SecretStr]] = None,
        file_path: Optional[Path] = None,
        file_bytes: Optional[bytes] = None,
    ):
        """
        Class responsible for parsing PDFs and returning raw text

        The page_range variable determines which pages are extracted.
        All pages are extracted by default.
        """
        self._passwords = passwords
        self.file_path = file_path
        self.file_bytes = file_bytes
        self.bank = bank

    @property
    def passwords(self):
        if not self._passwords:
            return self.bank.passwords
        return self._passwords

    @property
    def pdf_config(self):
        return self.bank.pdf_config

    @cached_property
    def page_range(self):
        return slice(*self.pdf_config.page_range)

    @cached_property
    def page_bbox(self):
        return self.pdf_config.page_bbox

    def open(self):
        """
        Opens and decrypts a PDF document
        """
        document = self.document

        if not document.is_encrypted:
            return document

        if not self.passwords:
            raise MissingPasswordError("No password found in PDF configuration")

        if not any(str(password) for password in self.passwords):
            raise MissingPasswordError("Password in PDF configuration is empty")

        if not isinstance(self.passwords, list):
            raise BadPasswordFormatError("Passwords should be stored in a list")

        if not all(isinstance(item, SecretStr) for item in self.passwords):
            raise BadPasswordFormatError("Passwords should be stored as SecretStr")

        for password in self.passwords:
            document.authenticate(password.get_secret_value())

            if not document.is_encrypted:
                logger.debug("Successfully authenticated with password")
                return document
        raise WrongPasswordError(f"Could not open document: {document.name}")

    @lru_cache
    def get_pages(self) -> list[PdfPage]:
        logger.debug("Extracting text from PDF")
        document: fitz.Document = self.open()

        num_pages = list(range(document.page_count))
        document.select(num_pages[self.page_range])

        if cropbox := self.page_bbox:
            logger.debug(
                "Will crop pages with crop box %s and remove vertical text", cropbox
            )
        for page in document:
            if self.page_bbox:
                page.set_cropbox(cropbox)
            page = self._remove_vertical_text(page)

        # certain statements require garbage collection, so that duplicate objects
        # do not cause pdftotext to fail due to missing xrefs/null values
        # however, setting `garbage=2` may cause issues with other statements
        # so an initial attempt should be made to run using `garbage=0`
        garbage_values = [0, 2]

        for garbage in garbage_values:
            try:
                pdf_byte_stream = BytesIO(document.tobytes(garbage=garbage))
                pdf = pdftotext.PDF(pdf_byte_stream, physical=True)
                return [PdfPage(page) for page in pdf]
            except pdftotext.Error:
                continue
        raise RuntimeError("Unable to retrieve pages")

    @cached_property
    def document(self) -> fitz.Document:
        """
        Returns a Python representation of a PDF document.
        """
        if not self.file_path and not self.file_bytes:
            raise RuntimeError("Either `file_path` or `file_bytes` must be passed")

        if self.file_path and self.file_bytes:
            raise RuntimeError(
                "Only one of `file_path` or `file_bytes` should be defined"
            )

        args = {"filename": self.file_path, "stream": self.file_bytes}
        return fitz.Document(**args)

    @staticmethod
    def _remove_vertical_text(page: fitz.Page):
        """Helper function to remove vertical text, based on writing direction (wdir).

        This helps avoid situations where the PDF is oddly parsed, due to vertical text
        inside the PDF.

        An example of vertical text breaking a transaction:
        ```
        'HEALTHY HARVEST CAFÉ SINGAPORE SG',
        'Co Reg No: 123456',
        '10 NOV 9.80',
        ```

        Note:
            The 'dir' key represents the tuple (cosine, sine) for the angle.
            If line["dir"] != (1, 0), the text of its spans is rotated.

        """
        for block in page.get_text("dict", flags=fitz.TEXTFLAGS_TEXT)["blocks"]:
            for line in block["lines"]:
                writing_direction = line["dir"]
                if writing_direction != (1, 0):
                    page.add_redact_annot(line["bbox"])
        page.apply_redactions(images=fitz.PDF_REDACT_IMAGE_NONE)
        return page



import re
from dataclasses import asdict
from enum import auto

from pydantic.dataclasses import dataclass
from strenum import StrEnum


# flake8: noqa
# pylint: disable=line-too-long
class DateFormats(StrEnum):
    D = r"(?:1|2|3|4|5|6|7|8|9|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|28|29|30|31)"
    DD = r"(?:01|02|03|04|05|06|07|08|09|10|11|12|13|14|15|16|17|18|19|20|21|22|23|24|25|26|27|28|29|30|31)"
    M = r"(?:1|2|3|4|5|6|7|8|9|10|11|12)"
    MM = r"(?:01|02|03|04|05|06|07|08|09|10|11|12)"
    MMM = r"(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)"
    MMMM = r"(?:January|February|March|April|May|June|July|August|September|October|November|December)"
    YY = r"(?:[2-5][0-9]\b)"
    YYYY = r"(?:20\d{2}\b)"


@dataclass
class DateRegexPatterns:
    """Holds date regex patterns used by the generic statement handler"""

    dd_mm: str = rf"\b({DateFormats.DD}[\/\-\s]{DateFormats.MM})"
    dd_mm_yy: str = (
        rf"\b({DateFormats.DD}[\/\-\s]{DateFormats.MM}[\/\-\s]{DateFormats.YY})"
    )
    dd_mmm: str = rf"\b({DateFormats.DD}[-\s]{DateFormats.MMM})"
    dd_mmm_yy: str = rf"\b({DateFormats.DD}[-\s]{DateFormats.MMM}[-\s]{DateFormats.YY})"
    dd_mmm_yyyy: str = (
        rf"\b({DateFormats.DD}[-\s]{DateFormats.MMM}[,\s]{{1,2}}{DateFormats.YYYY})"
    )
    dd_mm_yyyy: str = (
        rf"\b({DateFormats.DD}[\/\-\s]{DateFormats.MM}[\/\-\s]{DateFormats.YYYY})"
    )
    mmmm_dd_yyyy: str = (
        rf"\b({DateFormats.MMMM}\s{DateFormats.DD}[,\s]{{1,2}}{DateFormats.YYYY})"
    )
    mmm_dd: str = rf"\b({DateFormats.MMM}[-\s]{DateFormats.DD})"
    mmm_dd_yyyy: str = (
        rf"\b({DateFormats.MMM}[-\s]{DateFormats.DD}[,\s]{{1,2}}{DateFormats.YYYY})"
    )

    def as_pattern_dict(self):
        return {k: re.compile(v, re.IGNORECASE) for k, v in asdict(self).items()}


# pylint: disable=unused-argument,no-self-argument
class AutoEnum(StrEnum):
    """Generates lower case values for enums
    e.g. CITIBANK -> citibank
    """

    def _generate_next_value_(name: str, *_):  # type: ignore
        return name.lower()


class EntryType(AutoEnum):
    CREDIT = auto()
    DEBIT = auto()


class BankNames(AutoEnum):
    CITIBANK = auto()
    DBS = auto()
    HSBC = auto()
    MAYBANK = auto()
    OCBC = auto()
    STANDARD_CHARTERED = auto()


class InternalBankNames(AutoEnum):
    EXAMPLE = auto()
    GENERIC = auto()


class Columns(AutoEnum):
    AMOUNT = auto()
    DATE = auto()
    DESCRIPTION = auto()
    SUFFIX = auto()
    TRANSACTION_DATE = auto()


class SharedPatterns(StrEnum):
    """
    AMOUNT matches the following patterns:
    1,123.12 | 123.12 | (123.12) | ( 123.12) | 123.12 CR

    AMOUNT_EXTENDED is generally used for credit statements and to
    find statement balances, while AMOUNT_EXTENDED_WITHOUT_EOL is used
    for debit statements, since debit statements have the amount + balances
    on a single line, and we only want the amount.
    """

    COMMA_FORMAT = r"\d{1,3}(,\d{3})*\.\d*"
    ENCLOSED_COMMA_FORMAT = rf"\({COMMA_FORMAT}\s{{0,1}}\))"
    DEBIT_CREDIT_SUFFIX = r"(?P<suffix>CR\b|DR\b|\+|\-)?\s*"

    AMOUNT = rf"(?P<amount>{COMMA_FORMAT}|{ENCLOSED_COMMA_FORMAT}\s*"
    AMOUNT_EXTENDED_WITHOUT_EOL = AMOUNT + DEBIT_CREDIT_SUFFIX
    AMOUNT_EXTENDED = AMOUNT_EXTENDED_WITHOUT_EOL + r"$"

    BALANCE = rf"(?P<balance>{COMMA_FORMAT})?$"
    DESCRIPTION = r"(?P<description>.*?)\s+"
    TRANSACTION_DATE_ABBREVIATED_ALL_CAPS = r"(?P<transaction_date>\d{2}\s[A-Z]{3})\s+"
    TRANSACTION_DATE_ABBREVIATED_PROPER_CASE = (
        r"(?P<transaction_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"
    )
    POSTING_DATE_ABBREVIATED_PROPER = r"(?P<posting_date>\d{2}\s[A-Z]{1}[a-z]{2})\s+"


class StatementBalancePatterns(StrEnum):
    DBS = (
        r"(?P<description>PREVIOUS BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    CITIBANK = (
        r"(?P<description>BALANCE PREVIOUS STATEMENT?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    HSBC = (
        r"(?P<description>Previous Statement Balance?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    MAYBANK = (
        r"(?P<description>PREVIOUS STATEMENT BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    OCBC = (
        r"(?P<description>LAST MONTH'S BALANCE?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )
    STANDARD_CHARTERED = (
        r"(?P<description>BALANCE FROM PREVIOUS STATEMENT?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
    )


class CreditTransactionPatterns(StrEnum):
    DBS = (
        rf"(?P<transaction_date>{DateRegexPatterns.dd_mmm})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    CITIBANK = (
        rf"(?P<transaction_date>{DateRegexPatterns.dd_mmm})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    HSBC = (
        rf"(?P<posting_date>{DateRegexPatterns.dd_mmm})\s+"
        rf"(?P<transaction_date>{DateRegexPatterns.dd_mmm})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    MAYBANK = (
        rf"(?P<posting_date>{DateRegexPatterns.dd_mm})\s+"
        rf"(?P<transaction_date>{DateRegexPatterns.dd_mm})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    OCBC = (
        r"(?P<transaction_date>\d+/\d+)\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED
    )
    STANDARD_CHARTERED = (
        rf"(?P<transaction_date>{DateRegexPatterns.dd_mmm})\s+"
        rf"(?P<posting_date>{DateRegexPatterns.dd_mmm})\s+"
        + SharedPatterns.DESCRIPTION
        + r"(?:(?P<transaction_ref>Transaction\sRef\s\d+)?)\s+"
        + SharedPatterns.AMOUNT_EXTENDED
    )


class DebitTransactionPatterns(StrEnum):
    DBS = (
        rf"(?P<transaction_date>{DateRegexPatterns.dd_mmm})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
        + SharedPatterns.BALANCE
    )
    MAYBANK = (
        rf"(?P<transaction_date>{DateRegexPatterns.dd_mm_yy})\s+"
        + SharedPatterns.DESCRIPTION
        # remove *\s
        + SharedPatterns.AMOUNT[:-3]
        + r"(?P<suffix>\-|\+)\s+"
        + SharedPatterns.BALANCE
    )
    OCBC = (
        rf"(?P<transaction_date>{DateRegexPatterns.dd_mmm})\s+"
        rf"(?P<posting_date>{DateRegexPatterns.dd_mmm})\s+"
        + SharedPatterns.DESCRIPTION
        + SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
        + SharedPatterns.BALANCE
    )


@dataclass
class Identifier:
    pass


@dataclass
class EncryptionIdentifier(Identifier):
    pdf_version: float
    algorithm: int
    revision: int
    length: int
    permissions: int


@dataclass
class MetadataIdentifier(Identifier):
    title: str = ""
    author: str = ""
    subject: str = ""
    creator: str = ""
    producer: str = ""



import logging
from abc import ABC, abstractmethod
from typing import Optional

from pydantic import SecretStr

from monopoly.config import CreditStatementConfig, DebitStatementConfig, PdfConfig
from monopoly.constants import EncryptionIdentifier, MetadataIdentifier

logger = logging.getLogger(__name__)


class BankBase(ABC):
    """
    Abstract class to handle initialization of common variables
    that are shared between bank processor classes.

    Ensures consistency between bank classes.
    """

    credit_config: Optional[CreditStatementConfig] = None
    debit_config: Optional[DebitStatementConfig] = None
    # pdf_config defaults to an empty object if not overriden
    pdf_config: PdfConfig = PdfConfig()

    def __init__(self, generic=False):
        self.validate_config(generic)
        self.populate_pdf_config()

    def validate_config(self, generic: bool):
        # Basic validation to ensure required attributes are set
        if not generic:
            if self.credit_config is None and self.debit_config is None:
                raise NotImplementedError(
                    f"{self.__class__.__name__} "
                    "must implement either `credit_config` or `debit_config`"
                )

    def populate_pdf_config(self):
        # Ensure that PDF config always exists
        if not self.pdf_config:
            self.pdf_config = PdfConfig()
            logger.info(f"{self.__class__.__name__}: Using default `pdf_config`")

    @property
    @abstractmethod
    def identifiers(self) -> list[EncryptionIdentifier | MetadataIdentifier]:
        raise NotImplementedError("Identifiers must be defined")

    @property
    def passwords(self) -> Optional[list[SecretStr]]:
        if self.passwords:
            return self.passwords
        return None



import traceback
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Collection, Iterable, Optional, TypedDict

import click
from pydantic.dataclasses import dataclass
from tabulate import tabulate
from tqdm import tqdm

from monopoly.log import setup_logs


class TqdmSettings(TypedDict):
    """
    Stores settings for `tqdm`
    """

    total: int
    desc: str
    leave: bool
    delay: float
    ncols: int
    bar_format: str


@dataclass
class Result:
    """
    Stores information about processed bank statement
    """

    source_file_name: str
    target_file_name: Optional[str] = None
    error_info: Optional[dict[str, str]] = None


@dataclass
class Report:
    """
    A helper class for parsing and displaying processed bank statements.
    """

    results: list[Result]

    @property
    def processed_results(self) -> list[Result]:
        return sorted(
            [r for r in self.results if not r.error_info],
            key=lambda x: x.source_file_name,
        )

    @property
    def errored_results(self) -> list[Result]:
        return sorted(
            [r for r in self.results if r.error_info], key=lambda x: x.source_file_name
        )

    @property
    def number_processed(self) -> int:
        return len(self.processed_results)

    @property
    def number_errored(self) -> int:
        return len(self.errored_results)

    def display_report(self) -> None:
        """
        Parses all results, displaying the number of successfully
        processed statements and any errors.
        """
        if self.number_errored > 0:
            error_msg = (
                f"{self.number_errored} statement(s) had errors while processing"
            )
            click.echo(click.style(error_msg, fg="red", bold=True))
        if self.number_processed > 0:
            changed_msg = f"{self.number_processed} statement(s) processed"
            click.echo(click.style(changed_msg, bold=True))

        for res in self.errored_results:
            click.echo(
                click.style(
                    f"{res.source_file_name} -- "
                    f"{res.error_info['message'][:120]}",  # type: ignore
                    fg="red",
                )
            )

        for res in self.processed_results:
            click.echo(f"{res.source_file_name} -> {res.target_file_name}")


def process_statement(
    file: Path,
    output_directory: Optional[Path],
    print_df: bool,
    safety_check: bool = True,
) -> Optional[Result]:
    """
    Extracts, transforms, and loads transactions from bank statements.

    Parameters:
        file: The path to the bank statement file.
        output_directory: The directory to save the processed statement.
            Defaults to the parent directory of the input file if not provided.
        print_df: If True, the transformed DataFrame is printed to the console
            in a tabular format. No file is generated in this case.

    Returns:
        Optional[Result]: If print_df is False, returns a Result object containing
        information about the processed statement. If an error occurs during processing,
        returns a Result object with error information.
    """
    from monopoly.pipeline import Pipeline  # pylint: disable=import-outside-toplevel

    try:
        pipeline = Pipeline(file)
        statement = pipeline.extract(safety_check=safety_check)
        transactions = pipeline.transform(statement)

        if print_df:
            pprint_transactions(transactions, statement, file)
            # don't load to CSV if pprint
            return None

        # saves processed statements to the same directory by default
        if not output_directory:
            output_directory = file.parent

        output_file = pipeline.load(transactions, statement, output_directory)
        return Result(file.name, output_file.name)

    except Exception as err:  # pylint: disable=broad-exception-caught
        error_info = {
            "message": f"{type(err).__name__}: {str(err)}",
            "traceback": traceback.format_exc(),
        }
        return Result(file.name, error_info=error_info)


def pprint_transactions(transactions: list, statement, file: Path) -> None:
    """Prints transactions in a markdown style tabular format"""
    click.echo(f"{file.name}")
    transactions_as_dict = [transaction.as_raw_dict() for transaction in transactions]
    headers = {col: col for col in statement.columns}
    click.echo(
        tabulate(
            transactions_as_dict,
            headers=headers,
            tablefmt="psql",
            numalign="right",
        )
    )
    click.echo()


def run(
    input_files: Collection[Path],
    output_dir: Optional[Path] = None,
    pprint: bool = False,
    single_process: bool = False,
    safety_check: bool = True,
    **kwargs,
):
    """
    Process a collection of input files concurrently

    If any statements are processed successfully or encounter errors, a Report object
    is created and its display_report() method is called to provide a summary
    of the processing outcomes.
    """

    tqdm_settings: TqdmSettings = {
        "total": len(input_files),
        "desc": "Processing statements",
        "leave": False,
        "delay": 0.2,
        "ncols": 80,
        "bar_format": "{l_bar}{bar}| {n_fmt}/{total_fmt}",
    }

    if not single_process:
        with ProcessPoolExecutor() as executor:
            results = list(
                tqdm(
                    executor.map(
                        process_statement,
                        input_files,
                        [output_dir] * len(input_files),
                        [pprint] * len(input_files),
                        [safety_check] * len(input_files),
                    ),
                    **tqdm_settings,
                )
            )

    else:
        results = []
        for file in tqdm(input_files, **tqdm_settings):
            result = process_statement(file, output_dir, pprint, safety_check)
            results.append(result)

    if any(results):
        # filter out null values, for cases where print_df is True
        # and processing errors occur to avoid pydantic validation errors
        report = Report([res for res in results if res])
        report.display_report()


def get_statement_paths(files: Iterable[Path]) -> set[Path]:
    """
    Recursively collects paths to PDF files from a given collection of paths.
    """
    matched_files = set()
    for path in files:
        if path.is_file() and str(path).endswith(".pdf"):
            matched_files.add(path)

        if path.is_dir():
            matched_files |= get_statement_paths(path.iterdir())

    return matched_files


@click.command()
@click.version_option(package_name="monopoly-core")
@click.argument(
    "files",
    nargs=-1,
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
)
@click.option(
    "-o",
    "--output",
    "output_dir",
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
    help="Specify output folder.",
)
@click.option(
    "-p",
    "--pprint",
    is_flag=True,
    help="Print the processed statement(s).",
)
@click.option(
    "-s",
    "--single-process",
    is_flag=True,
    help=(
        "Runs `monopoly` in single-threaded mode, even when processing "
        "multiple files. Useful for debugging."
    ),
)
@click.option(
    "--safe/--nosafe",
    "safety_check",
    default=True,
    help=(
        "Determines whether to run the safety check or not. "
        "Runs the safety check by default."
    ),
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help=("Increases logging verbosity."),
)
@click.pass_context
@setup_logs
def monopoly(ctx: click.Context, files: list[Path], **kwargs):
    """
    Monopoly helps convert your bank statements from PDF to CSV.

    A file or directory can be passed in via the FILES argument
    """
    if files:
        matched_files = get_statement_paths(files)

        if matched_files:
            run(matched_files, **kwargs)

        else:
            click.echo(
                click.style("Could not find .pdf files", fg="yellow", bold=True),
                err=True,
            )
            ctx.exit(1)
    else:
        show_welcome_message()


def show_welcome_message():
    art = r"""
     __  __                               _
    |  \/  | ___  _ __   ___  _ __   ___ | |_   _
    | |\/| |/ _ \| '_ \ / _ \| '_ \ / _ \| | | | |
    | |  | | (_) | | | | (_) | |_) | (_) | | |_| |
    |_|  |_|\___/|_| |_|\___/| .__/ \___/|_|\__, |
                             |_|            |___/
    """

    click.echo(art)
    message = "Monopoly helps convert your bank statements from PDF to CSV"
    click.echo(message)
    commands = [
        (
            "monopoly .",
            "process all statements nested in current dir",
        ),
        (
            "monopoly path/to/file.pdf",
            "process a specific bank statement",
        ),
        (
            "monopoly . --output <dir>",
            "saves all results to a specific directory",
        ),
        (
            "monopoly . --pprint",
            "prints results instead of saving them",
        ),
        (
            "monopoly --help",
            "show more options and other usage information",
        ),
    ]
    margin = max(len(cmd) for cmd, _ in commands)
    for cmd, desc in commands:
        styled_cmd = click.style(
            text=f"{cmd}{' ' * (margin - len(cmd))}",
            fg="white",
            bg="bright_black",
            bold=True,
        )
        click.echo(message=f"    {styled_cmd} {desc}")
    click.echo()



import logging

from monopoly.config import CreditStatementConfig, DebitStatementConfig
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    DebitTransactionPatterns,
    MetadataIdentifier,
    StatementBalancePatterns,
)

from ..base import BankBase

logger = logging.getLogger(__name__)


class Maybank(BankBase):
    debit_config = DebitStatementConfig(
        bank_name=BankNames.MAYBANK,
        statement_date_pattern=r"(?:結單日期)[:\s]+(\d{2}\/\d{2}\/\d{2})",
        transaction_pattern=DebitTransactionPatterns.MAYBANK,
        has_withdraw_deposit_column=False,
        multiline_transactions=True,
    )

    credit_config = CreditStatementConfig(
        bank_name=BankNames.MAYBANK,
        statement_date_pattern=r"(\d{2}\s[A-Z]{3}\s\d{2})",
        transaction_pattern=CreditTransactionPatterns.MAYBANK,
        prev_balance_pattern=StatementBalancePatterns.MAYBANK,
        multiline_transactions=True,
    )

    identifiers = [
        MetadataIdentifier(
            author="Maybank2U.com",
            creator="Maybank2u.com",
            producer="iText",
        ),
    ]



import logging
import re
from functools import cached_property

from monopoly.constants import EntryType
from monopoly.statements.transaction import TransactionMatch

from .base import BaseStatement, SafetyCheckError

logger = logging.getLogger(__name__)


class DebitStatement(BaseStatement):
    """
    A dataclass representation of a debit statement
    """

    statement_type = EntryType.DEBIT

    @cached_property
    def debit_header(self):
        """Checks if the statement is a debit statement"""
        identifier = self.config.debit_statement_identifier
        for page in self.pages:
            for line in page.lines:
                if re.search(identifier, line):
                    return line.lower()
        return None

    def pre_process_match(
        self, transaction_match: TransactionMatch
    ) -> TransactionMatch:
        """
        Pre-processes transactions by adding a debit or credit suffix to the group dict
        """
        if self.config.has_withdraw_deposit_column:
            transaction_match.groupdict.suffix = self.get_debit_suffix(
                transaction_match
            )
        return transaction_match

    def get_debit_suffix(self, transaction_match: TransactionMatch) -> str | None:
        """
        Gets the accounting suffix for debit card statements

        Attempts to identify whether a transaction is a debit
        or credit entry based on the distance from the withdrawal
        or deposit columns.
        """
        if self.withdrawal_pos and self.deposit_pos:
            amount = transaction_match.groupdict.amount
            line: str = transaction_match.match.string
            start_pos = line.find(amount)
            # assume that numbers are right aligned
            end_pos = start_pos + len(amount) - 1
            withdrawal_diff = abs(end_pos - self.withdrawal_pos)
            deposit_diff = abs(end_pos - self.deposit_pos)
            if withdrawal_diff > deposit_diff:
                return "CR"
            return "DR"
        return transaction_match.groupdict.suffix

    @cached_property
    def withdrawal_pos(self) -> int | None:
        return self.get_column_pos("withdraw")

    @cached_property
    def deposit_pos(self) -> int | None:
        return self.get_column_pos("deposit")

    def get_column_pos(self, column_type: str) -> int | None:
        pattern = re.compile(rf"{column_type}[\w()$]*", re.IGNORECASE)
        match: re.Match | None = pattern.search(self.debit_header)
        if match:
            return self.get_header_pos(match.group())
        logger.warning(f"`{column_type}` column not found in header")
        return None

    def get_header_pos(self, column_name: str) -> int:
        """
        Returns position of the 'WITHDRAWAL' or 'DEPOSIT' header on a bank statement

        An assumption is made here that numbers are right aligned, meaning
        that if an amount matches with the end of the withdrawal string position,
        the item is in fact a withdrawal

        e.g.
        ```
        DATE         DESCRIPTION          WITHDRAWAL         DEPOSIT
        15 OCT       bill payment             322.07
        16 OCT       item                                     123.12
        ```
        """
        if self.debit_header:
            result = self.debit_header.find(column_name) + len(column_name)
            return result

        raise ValueError(f"Debit header {column_name} missing in {self.debit_header}")

    def perform_safety_check(self) -> bool:
        """
        Checks that debit and credit transaction sums
        exist as a number within the statement
        """
        transactions = self.transactions

        numbers = self.get_all_numbers_from_document()

        # add a zero, for cases where the debit statement
        # either is completely debit or credit transactions
        numbers.update([0])

        debit_amounts = [
            transaction.amount for transaction in transactions if transaction.amount > 0
        ]
        credit_amounts = [
            transaction.amount for transaction in transactions if transaction.amount < 0
        ]

        debit_sum = round(abs(sum(debit_amounts)), 2)
        credit_sum = round(abs(sum(credit_amounts)), 2)

        result = all([debit_sum in numbers, credit_sum in numbers])
        if not result:
            raise SafetyCheckError(self.failed_safety_message)

        return result



import logging
from functools import cached_property

from monopoly.banks import BankBase
from monopoly.config import CreditStatementConfig, DebitStatementConfig
from monopoly.constants import EntryType, InternalBankNames
from monopoly.handler import StatementHandler
from monopoly.pdf import PdfParser

from .generic import DatePatternAnalyzer

logger = logging.getLogger(__name__)


class GenericBank(BankBase):
    """
    Empty bank class with variables that can be populated by
    the `GenericStatementHandler` class
    """

    def __init__(self):
        super().__init__(generic=True)

    @property
    def identifiers(self):
        return None


class GenericStatementHandler(StatementHandler):
    def __init__(self, parser: PdfParser):
        pages = parser.get_pages()
        self.analyzer = DatePatternAnalyzer(pages)
        parser.bank.debit_config = self.debit_config
        parser.bank.credit_config = self.credit_config
        super().__init__(parser)

    @cached_property
    def debit_config(self):
        if self.statement_type == EntryType.DEBIT:
            logger.debug("Creating debit statement config")

            return DebitStatementConfig(
                bank_name=InternalBankNames.GENERIC,
                transaction_pattern=self.transaction_pattern,
                statement_date_pattern=self.statement_date_pattern,
                multiline_transactions=self.multiline_transactions,
                debit_statement_identifier=self.debit_statement_identifier,
            )
        return None

    @cached_property
    def credit_config(self):
        if self.statement_type == EntryType.CREDIT:
            logger.debug("Creating credit statement config")

            return CreditStatementConfig(
                bank_name=InternalBankNames.GENERIC,
                prev_balance_pattern=self.prev_balance_pattern,
                transaction_pattern=self.transaction_pattern,
                statement_date_pattern=self.statement_date_pattern,
                multiline_transactions=self.multiline_transactions,
            )
        return None

    @cached_property
    def transaction_pattern(self):
        return self.analyzer.create_transaction_pattern()

    @cached_property
    def statement_type(self):
        return self.analyzer.get_statement_type()

    @cached_property
    def statement_date_pattern(self):
        return self.analyzer.create_statement_date_pattern()

    @cached_property
    def multiline_transactions(self):
        return self.analyzer.check_if_multiline()

    @cached_property
    def debit_statement_identifier(self):
        return self.analyzer.get_debit_statement_header_line()

    @cached_property
    def prev_balance_pattern(self):
        return self.analyzer.create_previous_balance_regex()



from monopoly.pipeline import Pipeline


# pylint: disable=duplicate-code
def example():
    """Example showing how monopoly can be used to extract data from
    a single bank statement
    """
    pipeline = Pipeline(file_path="src/monopoly/examples/example_statement.pdf")

    # This runs pdftotext on the PDF and
    # extracts transactions as raw text
    statement = pipeline.extract()

    # Dates are converted into an ISO 8601 date format
    transactions = pipeline.transform(statement)

    # Parsed transactions writen to a CSV file in the "example" directory
    pipeline.load(
        transactions=transactions,
        statement=statement,
        output_directory="src/monopoly/examples",
    )


if __name__ == "__main__":
    example()



import logging

from monopoly.pdf import PdfParser
from monopoly.statements import CreditStatement, DebitStatement

logger = logging.getLogger(__name__)


class StatementHandler:
    """
    Retrieves statement information like transactions from the PDF,
    and identifies the statement as either a debit or credit statement
    based on the debit and credit config.
    """

    def __init__(self, parser: PdfParser):
        self.parser = parser
        self.bank = parser.bank
        self.statement = self.get_statement(self.parser)

    @property
    def transactions(self):
        return self.statement.transactions

    @property
    def statement_date(self):
        return self.statement.statement_date

    def perform_safety_check(self):
        self.statement.perform_safety_check()

    @classmethod
    def get_statement(cls, parser: PdfParser) -> CreditStatement | DebitStatement:
        bank = parser.bank
        debit_config, credit_config = bank.debit_config, bank.credit_config

        if debit_config:
            debit_statement = DebitStatement(parser, debit_config)
            # if we can find transactions using the debit config
            # assume that it is a debit statement
            try:
                if debit_statement.get_transactions():
                    return debit_statement
            except (ValueError, TypeError) as err:
                logger.debug(
                    "Could not parse with debit config due to error: %s",
                    err.__repr__(),
                )
            except Exception as err:
                logger.error(
                    "Unexpected error while parsing with debit config: %s",
                    err.__repr__(),
                )

        if not credit_config:
            raise RuntimeError("Missing credit config")

        # if it's not a debit statement, assume that it's a credit statement
        return CreditStatement(parser, credit_config)



import json
import re
from collections.abc import Mapping
from typing import Any, Optional

from pydantic import ConfigDict, Field, field_validator, model_validator
from pydantic.dataclasses import dataclass
from pydantic_core import ArgsKwargs

from monopoly.constants import Columns


class TransactionGroupDict(Mapping):
    """
    Helper class that holds the group dict from a match. This
    class acts as a convenient wrapper to allow unpacking into the
    `Transaction` class.

    A transaction should at minimum always have a description and amount

    In cases where we parse a previous balance line, the date might not exist
    e.g.
    `          LAST MONTH'S BALANCE         6,321`
    `01 OCT    ValueVille                  123.12`
    """

    def __init__(
        self,
        description: str,
        amount: str,
        transaction_date: Optional[str] = None,
        suffix: Optional[str] = None,
        **_,
    ):
        self.transaction_date = transaction_date
        self.amount = amount
        self.description = description
        self.suffix = suffix

    def __getitem__(self, x):
        return self.__dict__[x]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def asdict(self):
        return self.__dict__

    def __str__(self):
        return str(self.__dict__)

    def __repr__(self):
        return (
            "<TransactionGroupDict object; "
            f"{Columns.TRANSACTION_DATE}={self.transaction_date}, "
            f"{Columns.AMOUNT}={self.amount}, "
            f"{Columns.DESCRIPTION}={self.description}>"
        )


@dataclass(config=ConfigDict(arbitrary_types_allowed=True))
class TransactionMatch:
    groupdict: TransactionGroupDict
    match: re.Match

    def span(self):
        return self.match.span()


@dataclass
class Transaction:
    """
    Holds transaction data, validates the data, and
    performs various coercions like removing whitespaces
    and commas.

    Reassigns 'transaction_date' to 'date'
    """

    description: str
    amount: float
    date: str = Field(alias="transaction_date")
    suffix: Optional[str] = None

    def as_raw_dict(self, show_suffix=False):
        """Returns stringified dictionary version of the transaction"""
        items = {
            Columns.DATE.value: self.date,
            Columns.DESCRIPTION.value: self.description,
            Columns.AMOUNT.value: str(self.amount),
        }
        if show_suffix:
            items[Columns.SUFFIX] = self.suffix
        return items

    @field_validator("description", mode="after")
    def remove_extra_whitespace(cls, value: str) -> str:
        return " ".join(value.split())

    @field_validator(Columns.AMOUNT, mode="before")
    def prepare_amount_for_float_coercion(cls, amount: str) -> str:
        """
        Replaces commas, whitespaces and parentheses in string representation of floats
        e.g.
            1,234.00 -> 1234.00
            (-10.00) -> -10.00
            (-1.56 ) -> -1.56
        """
        if isinstance(amount, str):
            return re.sub(r"[,)(\s]", "", amount)
        return amount

    # pylint: disable=bad-classmethod-argument
    @model_validator(mode="before")  # type: ignore
    def treat_parenthesis_enclosure_as_credit(self: ArgsKwargs | Any) -> "ArgsKwargs":
        """
        Treat amounts enclosed by parentheses (e.g. cashback) as a credit entry
        """
        if self.kwargs:
            amount: str = self.kwargs[Columns.AMOUNT]
            if isinstance(amount, str):
                if amount.startswith("(") and amount.endswith(")"):
                    self.kwargs[Columns.SUFFIX] = "CR"
        return self

    @model_validator(mode="after")
    def convert_credit_amount_to_negative(self: "Transaction") -> "Transaction":
        """
        Converts transactions with a suffix of "CR" or "+" to positive
        """
        if self.suffix in ("CR", "+"):
            self.amount = abs(self.amount)

        else:
            self.amount = -abs(self.amount)
        return self

    def __str__(self):
        return json.dumps(self.as_raw_dict(show_suffix=True))



import logging
from dataclasses import dataclass, field
from functools import cached_property
from io import BytesIO
from pathlib import Path
from typing import Any, Optional

import fitz

from monopoly.banks import detect_bank
from monopoly.constants import EncryptionIdentifier, Identifier, MetadataIdentifier

logger = logging.getLogger(__name__)


@dataclass
class EncryptDict:
    """Stores encryption dictionary of a PDF"""

    raw_encrypt_dict: dict
    pdf_version: str
    algorithm: int = field(init=False)
    length: int = field(init=False)
    permissions: int = field(init=False)
    revision: int = field(init=False)

    def __post_init__(self):
        self.pdf_version = float(self.pdf_version[-3:])
        self.algorithm = int(self.raw_encrypt_dict.get("V"))
        self.length = int(self.raw_encrypt_dict.get("Length"))
        self.permissions = int(self.raw_encrypt_dict.get("P"))
        self.revision = int(self.raw_encrypt_dict.get("R"))


class MetadataAnalyzer:
    def __init__(
        self, file_path: Optional[Path] = None, file_bytes: Optional[bytes] = None
    ):
        if file_path and file_bytes:
            raise ValueError(
                "Only one of `file_path` or `file_bytes` should be provided"
            )

        self.file_path = file_path
        self.file_bytes = file_bytes

    @property
    def bank(self):
        return detect_bank(self.metadata_items)

    @property
    def document(self):
        args = {"filename": self.file_path, "stream": self.file_bytes}
        return fitz.Document(**args)

    @property
    def metadata_items(self) -> list[Any]:
        """
        Retrieves encryption and metadata identifiers from a bank statement PDF
        """
        identifiers: list[Identifier] = []
        if encrypt_dict := self.encrypt_dict:
            encryption_identifier = EncryptionIdentifier(
                float(encrypt_dict.pdf_version),
                encrypt_dict.algorithm,
                encrypt_dict.revision,
                encrypt_dict.length,
                encrypt_dict.permissions,
            )
            logger.debug("Found encryption identifier: %s", encryption_identifier)
            identifiers.append(encryption_identifier)

        if metadata := self.document.metadata:
            metadata_identifier = MetadataIdentifier(**metadata)
            logger.debug("Found metadata identifier: %s", metadata_identifier)
            identifiers.append(metadata_identifier)

        if not identifiers:
            raise ValueError("Could not get identifier")

        return identifiers

    @cached_property
    def encrypt_dict(self) -> EncryptDict | None:
        stream = self.get_doc_byte_stream()
        pdf_version_string = stream.read(8).decode("utf-8", "backslashreplace")

        if self.document.is_encrypted:
            raw_encrypt_dict = self.get_raw_encrypt_dict(self.document)
            encrypt_dict = EncryptDict(raw_encrypt_dict, pdf_version_string)
            return encrypt_dict
        return None

    @staticmethod
    def get_raw_encrypt_dict(doc: fitz.Document) -> dict:
        """
        Helper function to extract the PDF encryption dictionary, since
        `fitz` doesn't provide it
        """
        encrypt_metadata = {}
        pdf_object, value = doc.xref_get_key(-1, "Encrypt")
        if pdf_object != "xref":
            pass  # PDF has no metadata
        else:
            xref = int(value.replace("0 R", ""))  # extract the metadata xref
            for key in doc.xref_get_keys(xref):
                encrypt_metadata[key] = doc.xref_get_key(xref, key)[1]
        return encrypt_metadata

    def get_doc_byte_stream(self) -> BytesIO:
        if self.file_path:
            with open(self.file_path, "rb") as file:
                stream = BytesIO(file.read())
            return stream

        if self.file_bytes:
            return BytesIO(self.file_bytes)

        raise RuntimeError("Unable to create document")



import logging
import re
from abc import ABC
from datetime import datetime
from functools import cached_property, lru_cache
from pathlib import Path

from dateparser import parse

from monopoly.config import CreditStatementConfig, DebitStatementConfig
from monopoly.constants import Columns, SharedPatterns
from monopoly.pdf import PdfParser
from monopoly.statements.transaction import (
    Transaction,
    TransactionGroupDict,
    TransactionMatch,
)

# pylint: disable=bad-classmethod-argument
logger = logging.getLogger(__name__)


class BaseStatement(ABC):
    """
    A dataclass representation of a bank statement, containing
    the PDF pages (their raw text representation in a `list`),
    and specific bank config.
    """

    def __init__(
        self,
        parser: PdfParser,
        config: CreditStatementConfig | DebitStatementConfig,
    ):
        self.pages = parser.get_pages()
        self.config = config
        self.columns: list[str] = [
            Columns.DATE,
            Columns.DESCRIPTION,
            Columns.AMOUNT,
        ]
        self.parser = parser
        self.document = parser.document

    @property
    def bank(self):
        return self.parser.bank

    @property
    def pattern(self):
        pattern = self.config.transaction_pattern
        if not isinstance(self.config.transaction_pattern, re.Pattern):
            pattern = re.compile(self.config.transaction_pattern, re.IGNORECASE)
        return pattern

    @lru_cache
    def get_transactions(self) -> list[Transaction] | None:
        transactions: list[Transaction] = []

        for page in self.pages:
            for idx, line in enumerate(page.lines):
                if match := self.pattern.search(line):
                    groupdict = TransactionGroupDict(**match.groupdict())
                    transaction_match = TransactionMatch(groupdict, match)
                    pre_processed_match = self.pre_process_match(transaction_match)
                    processed_match = self.process_match(
                        match=pre_processed_match,
                        line=line,
                        lines=page.lines,
                        idx=idx,
                    )
                    transaction = Transaction(**processed_match.groupdict)
                    transactions.append(transaction)

        if not transactions:
            return None

        post_processed_transactions = self.post_process_transactions(transactions)
        return post_processed_transactions

    def pre_process_match(
        self, transaction_match: TransactionMatch
    ) -> TransactionMatch:
        return transaction_match

    def post_process_transactions(
        self, transactions: list[Transaction]
    ) -> list[Transaction]:
        return transactions

    def process_match(
        self,
        match: TransactionMatch,
        line: str,
        lines: list[str],
        idx: int,
    ) -> TransactionMatch:
        if self.config.multiline_transactions and idx < len(lines) - 1:
            multiline_description = self.get_multiline_description(
                match.groupdict.description, line, lines, idx
            )
            match.groupdict.description = multiline_description
        return match

    def get_multiline_description(
        self,
        description: str,
        current_line: str,
        lines: list[str],
        idx: int,
    ) -> str:
        """Checks if a transaction description spans multiple lines, and
        tries to combine them into a single string"""
        description_pos = current_line.find(description)
        for next_line in lines[idx + 1 :]:  # noqa: E203
            # if next line is blank, don't add the description
            if not next_line:
                break

            # if transaction found on next line then break
            if self.pattern.search(next_line):
                break

            # don't process line if the description across both lines
            # doesn't align with a margin of three spaces
            next_line_text = next_line.strip()
            next_line_start_pos = next_line.find(next_line_text.split(" ")[0])

            if abs(description_pos - next_line_start_pos) > 3:
                break

            # if there's an amount in the next line, and it's further than
            # 20 spaces away, we assume that this isn't part of the description
            # and is a footer line like "Total" or "Balance Carried Forward"
            next_line_words = re.search(r"\s[A-Za-z]+", next_line)
            next_line_numbers = re.search(SharedPatterns.AMOUNT, next_line)

            if next_line_words and next_line_numbers:
                _, words_end_pos = next_line_words.span()
                numbers_start_pos, _ = next_line_numbers.span()

                if (
                    numbers_start_pos > words_end_pos
                    and numbers_start_pos - words_end_pos > 20
                ):
                    break

            description += " " + next_line

        return description

    @property
    def failed_safety_message(self) -> str:
        return (
            f"Safety check for {Path(self.document.name).stem} failed - "
            "transactions may be inaccurate"
        )

    @cached_property
    def transactions(self):
        return self.get_transactions()

    @cached_property
    def statement_date(self) -> datetime:
        for page in self.pages:
            for line in page.lines:
                if match := re.search(self.config.statement_date_pattern, line):
                    statement_date = parse(
                        date_string=match.group(1),
                        settings=self.config.statement_date_order.settings,
                    )
                    if statement_date:
                        return statement_date
        raise ValueError("Statement date not found")

    @staticmethod
    def get_decimal_numbers(lines: list[str]) -> set[float]:
        """Returns all decimal numbers in a statement. This is used
        to perform a safety check, to make sure no transactions have been missed"""

        number_pattern = re.compile(r"[\d.,]+")
        decimal_pattern = re.compile(r"\d+\.\d+$")
        numbers: set[str] = set()
        for line in lines:
            numbers.update(number_pattern.findall(line))
            numbers = {number.replace(",", "") for number in numbers}
            decimal_numbers = {
                float(number) for number in numbers if decimal_pattern.match(number)
            }
        return decimal_numbers

    def perform_safety_check(self) -> bool:
        """Placeholder for perform_safety_check method, which should
        exist in any child class of Statement"""
        raise NotImplementedError(
            "Subclasses must implement perform_safety_check method"
        )

    def get_all_numbers_from_document(self) -> set:
        """
        Iterates over each page in a statement, and retrieves
        all decimal numbers in each page. This is used by child classes with
        implementations of perform_safety_check()
        """
        numbers = set()

        for page in self.pages:
            lines = page.lines
            numbers.update(self.get_decimal_numbers(lines))
        numbers.add(self.get_subtotal_sum())
        return numbers

    def get_subtotal_sum(self):
        """
        Retrieves the subtotals from a document, and calculates the total.
        Useful for statements that don't give a total figure over
        several cards/months in a single statement.
        """
        subtotal_pattern = re.compile(
            rf"(?:sub\stotal.*?)\s+{SharedPatterns.AMOUNT}", re.IGNORECASE
        )
        subtotals: list[str] = []
        for page in self.pages:
            for line in page.lines:
                if match := subtotal_pattern.search(line):
                    subtotals.append(match.groupdict()[Columns.AMOUNT])
        cleaned_subtotals = [float(amount.replace(",", "")) for amount in subtotals]
        return sum(cleaned_subtotals)


class SafetyCheckError(Exception):
    """Raise when safety check fails"""



import logging
import re
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property, lru_cache

from dateparser import parse

from monopoly.constants import DateRegexPatterns, EntryType, SharedPatterns
from monopoly.pdf import PdfPage

logger = logging.getLogger(__name__)


@dataclass
class DateMatch:
    """Helper class that holds information about a match, including the span
    (start pos and end pos), the parsed date, the raw date, the line and
    page number, as well as the entire line that was matched
    """

    span: tuple[int, int]
    parsed_date: datetime
    raw_date: str
    page_number: int
    line_number: int
    line: str


class DatePatternAnalyzer:
    def __init__(self, pages: list[PdfPage]):
        self.pages = pages
        self.date_regex_patterns: dict[str, re.Pattern] = (
            DateRegexPatterns().as_pattern_dict()
        )
        self.patterns_with_date_matches = self.get_patterns_with_date_matches()
        self.pattern_spans_mapping = self.find_tuples_with_highest_occurrence(
            self.unique_matches_by_pattern
        )
        self.filtered_lines_with_dates = self.get_filtered_lines_with_dates(
            self.pattern_spans_mapping
        )

    @cached_property
    def unique_matches_by_pattern(self) -> dict:
        return self.count_unique_date_matches()

    @cached_property
    def lines_before_first_transaction(self) -> list[str]:
        first_transaction_line: DateMatch = self.filtered_lines_with_dates[0]
        first_page = self.pages[first_transaction_line.page_number].lines
        return list(reversed(first_page[: first_transaction_line.line_number]))

    def get_patterns_with_date_matches(self) -> dict[str, list[DateMatch]]:
        """Retrieves all possible transactions from every page, based on the assumption
        that a transaction line will always start with a date"""
        results: dict[str, list] = {}
        results = {
            pattern_name: self._find_matches(pattern, pattern_name, self.pages)
            for pattern_name, pattern in self.date_regex_patterns.items()
        }

        # filter out patterns with no matches
        return {k: v for k, v in results.items() if v}

    def _find_matches(
        self, pattern: re.Pattern, pattern_name: str, pages: list[PdfPage]
    ) -> list[DateMatch]:
        matches = []
        logger.debug("Searching for date matches for pattern %s", pattern_name)
        for page_num, page in enumerate(pages):
            for line_num, line in enumerate(page.lines):
                matches.extend(
                    self._extract_date_matches(pattern, line, page_num, line_num)
                )
        return matches

    def _extract_date_matches(
        self, pattern: re.Pattern, line: str, page_num: int, line_num: int
    ) -> list[DateMatch]:
        matches = []
        for date_match in pattern.finditer(line):
            raw_date = date_match.group()
            parsed_date = parse(raw_date)
            if parsed_date:
                matches.append(
                    DateMatch(
                        span=(date_match.start(), date_match.end()),
                        parsed_date=parsed_date,
                        raw_date=raw_date,
                        page_number=page_num,
                        line_number=line_num,
                        line=date_match.string,
                    )
                )
                logger.debug(
                    "Date match found at page %s line %s: raw_date='%s' parsed_date='%s'",
                    page_num,
                    line_num,
                    raw_date,
                    parsed_date,
                )
        return matches

    def count_unique_date_matches(self) -> dict:
        match_dict: dict = {}

        for pattern_name, matches in self.patterns_with_date_matches.items():
            match_dict[pattern_name] = {}

            unique_matches = set()
            for match in matches:
                unique_matches.add(match.span)

            sorted_unique_matches = sorted(unique_matches)

            for match in sorted_unique_matches:  # type: ignore
                match_dict[pattern_name][match] = 0

            # count number of times each tuple occurs
            for match in matches:
                match_tuple = match.span
                if match_tuple in match_dict[pattern_name]:
                    match_dict[pattern_name][match_tuple] += 1

        logger.debug("Unique matches: %s", match_dict)
        return match_dict

    @staticmethod
    def find_tuples_with_highest_occurrence(
        unique_date_matches: dict[str, dict[tuple, int]]
    ) -> dict:
        """
        Finds one or more tuples with the highest occurrence over several lines.
        For example, a line like '01 Oct  <description>' has a span of (0, 6) - we then
        check for repeats of this span, and sum it over all lines.

        If there are two tuples with the same number of occurrences e.g. (0, 6) has 17
        occurrences and (11, 17) has 17 occurrences, we return both
        tuples since there is a strong likelihood that one is a transaction date,
        and the other is a posting date.

        Patterns ending with "yy" are analyzed first, to avoid cases where we have
        a tiebreaker with dd_mm and dd_mm_yy variants.
        """
        span_occurrences = 0
        most_common_pattern = ""
        most_common_tuples = []

        # Sort patterns so that those ending with "yy" come first
        sorted_patterns = sorted(
            unique_date_matches.keys(), key=lambda p: not p.endswith("yy")
        )

        for pattern in sorted_patterns:
            spans_dict = unique_date_matches[pattern]
            max_span_occurrences = max(unique_date_matches[pattern].values())

            if max_span_occurrences > span_occurrences:
                tuples = [k for k, v in spans_dict.items() if v == max_span_occurrences]
                span_occurrences = max_span_occurrences
                most_common_pattern = pattern
                most_common_tuples = tuples

        logger.debug(
            "Pattern %s matched %s times at span(s) %s",
            most_common_pattern,
            span_occurrences,
            most_common_tuples,
        )

        # If the another span tuple pair occurs
        # at a rate of at least 0.5x the most common pair,
        # add it to the most_common_tuples list
        # this adds tolerance for tuples that may be misaligned across pages
        # e.g. {(3, 8): 7, (4, 9): 16, (23, 28): 23}
        if len(most_common_tuples) == 1:
            del unique_date_matches[most_common_pattern][most_common_tuples[0]]
            for key in unique_date_matches[most_common_pattern].keys():
                if (
                    unique_date_matches[most_common_pattern][key]
                    > span_occurrences * 0.5
                ):
                    most_common_tuples.append(key)
                    logger.debug("Adding additional span: %s", key)
                    break
        logger.debug(
            "Most common pattern / tuples: %s",
            {most_common_pattern: most_common_tuples},
        )
        return {most_common_pattern: most_common_tuples}

    @staticmethod
    def get_items_with_same_page_and_line(items: list[DateMatch]) -> list[tuple]:
        """Groups date items with the same line and page number"""
        grouped_items: defaultdict[str, list] = defaultdict(list)
        for item in items:
            key = f"{item.page_number}-{item.line_number}"
            grouped_items[key].append(item.parsed_date)

        # Convert the dictionary into a list of tuples
        result = [(key, *values) for key, values in grouped_items.items()]
        return result

    def get_filtered_lines_with_dates(
        self, pattern_spans_mapping: dict
    ) -> list[DateMatch]:
        """
        Helper function to retrieve lines that are the most likely to contain
        transactions, based on the transaction pattern and spans from the
        `find_tuples_with_highest_occurrence()` function
        """
        pattern, spans = next(iter(pattern_spans_mapping.items()))
        filtered_lines_with_dates = []
        for line in self.patterns_with_date_matches[pattern]:
            if line.span in spans:
                filtered_lines_with_dates.append(line)

        if not filtered_lines_with_dates:
            raise ValueError(f"Span(s) {list(spans)} not found in lines")
        return filtered_lines_with_dates

    @lru_cache
    def is_transaction_date_first(self) -> bool:
        """
        Check that in cases where date_1 > date_2, the first date is
        identified as the posting date
        """
        # only focus on our main transaction pattern
        lines_with_multiple_dates = self.get_items_with_same_page_and_line(
            self.filtered_lines_with_dates
        )
        for line in lines_with_multiple_dates:
            logger.debug("Checking dates in line: %s", line)
            if len(line) > 2:
                _, date_1, date_2 = line
                if date_1 and date_2:
                    if date_1 != date_2:
                        logger.debug(
                            "%s != %s, returning %s", date_1, date_2, date_1 < date_2
                        )
                        return date_1 < date_2

        logger.warning(
            "Could not identify difference between transaction and posting date"
        )
        return True

    @lru_cache
    def create_transaction_pattern(
        self,
    ) -> str:
        """
        Create a regex pattern that will be used for date parsing
        by the generic statement handler.
        """
        pattern = next(iter(self.pattern_spans_mapping))
        logger.debug("Creating date pattern for %s", pattern)
        transaction_date_regex = self.date_regex_patterns[pattern].pattern
        date_regex = rf"(?P<transaction_date>{transaction_date_regex})\s+"
        spans = self.pattern_spans_mapping[pattern]

        if len(spans) > 2:
            raise ValueError("More than two date tuples found in statement")

        pattern = date_regex

        # if there are two spans, we attempt to determine which is the
        # posting date, and which is the transaction date
        if len(spans) == 2:
            logger.debug("Creating group for posting date")
            posting_date = f"(?P<posting_date>{transaction_date_regex})"
            if self.is_transaction_date_first():
                pattern = date_regex + rf"{posting_date}\s+"
            else:
                pattern = rf"{posting_date}\s+" + date_regex

        pattern += SharedPatterns.DESCRIPTION

        # if the statement is debit-type statement, we add the amount + balance patterns
        if self.get_statement_type() == EntryType.DEBIT:
            pattern += SharedPatterns.AMOUNT_EXTENDED_WITHOUT_EOL
            pattern += SharedPatterns.BALANCE

        if self.get_statement_type() == EntryType.CREDIT:
            pattern += SharedPatterns.AMOUNT_EXTENDED

        return pattern

    @lru_cache
    def create_statement_date_pattern(self) -> str:
        """
        Creates a regex pattern for the statement date based on the first statement
        date.
        """
        lines_with_yyyy_dates = {
            pattern: lines
            for pattern, lines in self.patterns_with_date_matches.items()
            if pattern.endswith("yy")
        }

        if not lines_with_yyyy_dates:
            raise RuntimeError(
                "No lines with `yy` or `yyyy` dates "
                "- unable to create statement date pattern"
            )

        # try yyyy patterns first before yy
        sorted_patterns = sorted(
            lines_with_yyyy_dates.keys(), key=lambda p: not p.endswith("yyyy")
        )

        unsorted_lines = []

        for pattern in sorted_patterns:
            lines = lines_with_yyyy_dates[pattern]
            for line in lines:
                unsorted_lines.append((pattern, line))

        sorted_lines: list[tuple[str, DateMatch]] = sorted(
            unsorted_lines, key=lambda x: (x[1].page_number, x[1].line_number)
        )

        _, date_match = sorted_lines[0]
        statement_date_pattern = f"({date_match.raw_date})"
        logger.debug("Found statement date pattern: %s", statement_date_pattern)

        # assume the statement date is the first parsable YYYY date
        return statement_date_pattern

    @lru_cache
    def get_statement_type(self) -> str:
        """
        Since debit and credit statements need to be handled differently,
        we use the average number of matches per line as a way to determine
        if a statement is a debit or credit statement.

        For example, a debit statement transaction line might be something like:
        `01 OCT   ValueVille    123.12     2,000.00`
        where the first number is either a deposit/withdrawal,
        and the next number is a balance.
        """
        # match up to amounts like `123.12     2,000.00` while ignoring words between
        amount_pattern = re.compile(
            r"(?P<amount>\d{1,3}(,\d{3})*(\.\d+)\d{1,3}(,\d{3})*(\.\d+)?)"
        )

        def get_num_amounts_per_line(lines: list[DateMatch]) -> float:
            matches = [
                result
                for line in lines
                for result in amount_pattern.finditer(line.line)
            ]
            return len(matches) / len(lines)

        average_matches_per_line = get_num_amounts_per_line(
            self.filtered_lines_with_dates
        )
        # this technically should be 1.0, but 1.1 provides some
        # slight tolerance for error
        if average_matches_per_line > 1.1:
            logger.debug(
                "Found %s amounts per line - assuming debit statement",
                average_matches_per_line,
            )
            return EntryType.DEBIT

        return EntryType.CREDIT

    @lru_cache
    def get_debit_statement_header_line(self) -> str:
        header_pattern = re.compile(r"\b(date.*$)", re.IGNORECASE)
        for line in self.lines_before_first_transaction:
            if line:
                # assume that the header always starts with `date` or `DATE`
                result = header_pattern.search(line)

                if result:
                    escaped_result = f"({re.escape(result.string)})"
                    logger.debug("Found header statement: %s", escaped_result)
                    return escaped_result

        raise RuntimeError("Could not find debit statement header line")

    @lru_cache
    def check_if_multiline(self) -> bool:
        """Checks if the statement should be treated as a multiline statement,
        where the description is split across multiple lines
        """
        line_numbers = [item.line_number for item in self.filtered_lines_with_dates]
        line_distance = [
            line_numbers[i] - line_numbers[i - 1] for i in range(1, len(line_numbers))
        ]
        # remove negative numbers, since this indicates that the transaction is
        # from the next page
        line_distance = [i for i in line_distance if i > 0]

        # if there's more than two lines between each transaction on average
        # we assume that this is a multiline statement
        average_line_distance = sum(line_distance) / len(line_distance)
        logger.debug(
            "Assuming multiline statement: average distance between transaction lines: %s",
            average_line_distance,
        )
        return average_line_distance > 2

    @lru_cache
    def create_previous_balance_regex(self) -> str | None:
        """Helper function to check for a previous balance line items.
        Makes the assumption that the previous balance line item, if it
        exists, will be the line before the first line item with a date.

        e.g.
        ```
        PREVIOUS BALANCE                       6,492.54
        05 OCT PAYMENT - DBS INTERNET/WIRELESS 6,492.54 CR
        ```
        """
        page_num, line_num = self.get_first_transaction_location()

        # grab the first five lines before the first transaction
        page = self.pages[page_num]
        lines_before_first_transaction = list(
            reversed(page.lines[line_num - 5 : line_num])
        )

        balance_pattern = re.compile(SharedPatterns.DESCRIPTION + SharedPatterns.AMOUNT)

        for line in lines_before_first_transaction:
            balance_match = balance_pattern.search(line)
            if balance_match:
                logger.debug("Balance match found")
                if word_match := re.search(r"(\b[a-zA-Z'\s]+)", balance_match.string):
                    words = word_match[0].strip()
                    pattern = (
                        SharedPatterns.DESCRIPTION.replace(".*?", words)
                        + SharedPatterns.AMOUNT
                    )
                    logger.debug("Found words, generated pattern %s", pattern)
                    return pattern
        return None

    @lru_cache
    def get_first_transaction_location(self):
        # uses the transaction pattern to find the first transaction
        pattern = re.compile(self.create_transaction_pattern(), re.IGNORECASE)

        for page_num, page in enumerate(self.pages):
            for line_num, line in enumerate(page.lines):
                match = pattern.search(line)
                if match:
                    logger.debug(
                        "Found first transaction at page %s line %s", page_num, line_num
                    )
                    return page_num, line_num
        raise RuntimeError("Could not find first transaction")



import logging

from monopoly.config import CreditStatementConfig, PdfConfig, passwords
from monopoly.constants import (
    BankNames,
    CreditTransactionPatterns,
    EncryptionIdentifier,
    MetadataIdentifier,
    StatementBalancePatterns,
)

from ..base import BankBase

logger = logging.getLogger(__name__)


class Hsbc(BankBase):
    credit_config = CreditStatementConfig(
        bank_name=BankNames.HSBC,
        statement_date_pattern=r"Statement From .* to (\d{2}\s[A-Z]{3}\s\d{4})",
        prev_balance_pattern=StatementBalancePatterns.HSBC,
        transaction_pattern=CreditTransactionPatterns.HSBC,
        multiline_transactions=True,
    )

    pdf_config = PdfConfig(
        page_bbox=(0, 0, 379, 842),
    )

    identifiers = [
        EncryptionIdentifier(
            pdf_version=1.6, algorithm=4, revision=4, length=128, permissions=-1804
        ),
        MetadataIdentifier(
            title="PRJ_BEAGLE_ST_CNS_SGH_APP_Orchid",
            creator="OpenText Exstream",
        ),
    ]

    passwords = passwords.hsbc_pdf_passwords
